""" 
Action for PloneFormGen that helps you getpaid.
"""

__author__  = 'Daniel Holth <dholth@fastmail.fm>'
__docformat__ = 'plaintext'

import logging

from AccessControl import ClassSecurityInfo
from Acquisition import aq_parent
from zope.interface import providedBy
from zope import component
import zope.component

from AccessControl import getSecurityManager
from Products.Archetypes import atapi
from Products.Archetypes.public import StringField, SelectionWidget, \
    DisplayList, Schema, StringWidget, BooleanField, BooleanWidget
from Products.ATContentTypes.content.schemata import finalizeATCTSchema
from Products.CMFCore.permissions import ModifyPortalContent
from Products.CMFCore.utils import getToolByName

from Products.DataGridField import DataGridField, DataGridWidget
from Products.DataGridField.SelectColumn import SelectColumn
from Products.DataGridField.FixedColumn import FixedColumn
from Products.DataGridField.DataGridField import FixedRow

from Products.PloneFormGen.interfaces import IPloneFormGenField
from getpaid.core import interfaces as GPInterfaces
import getpaid.core

from Products.PloneFormGen.content.actionAdapter import \
    FormActionAdapter, FormAdapterSchema
from Products.PloneFormGen.config import FORM_ERROR_MARKER

from getpaid.formgen.config import PROJECTNAME
from zope.app import zapi
from getpaid.formgen.content.checkout import MakePaymentProcess
from Products.PloneGetPaid.interfaces import IGetPaidManagementOptions, IVariableAmountDonatableMarker
from Products.PloneGetPaid import sessions

logger = logging.getLogger("PloneFormGen")

schema = FormAdapterSchema.copy() + Schema((
    StringField('GPFieldsetType',
                searchable=False,
                required=False,
                mutator='setGPTemplate',
                widget=SelectionWidget(
                       label=u'Get Paid Form Template',
                       i18n_domain = "getpaidpfgadapter",
                       label_msgid = "label_getpaid_template",
                       ),
                vocabulary='getAvailableGetPaidForms'
                ),
                
    BooleanField('useFormAsContinueDestination',
        required=0,
        searchable=0,
        default='0',
        widget=BooleanWidget(
            label="Use form as 'Continue Shopping' destination.",
            description="""
                Check this to instruct the getpaid cart
                to navigate the user back to this form
                if they click 'Continue Shopping' while
                viewing their cart.
                """,
            label_msgid = "label_continue_shopping_text",
            description_msgid = "help_continue_shopping_text",
            i18n_domain = "getpaidpfgadapter",
            ),
        ),

    DataGridField('payablesMap',
         searchable=False,
         required=True,
         columns=('field_path', 'form_field', 'payable_path'),
         fixed_rows = "generateFormFieldRows",
         allow_delete = False,
         allow_insert = False,
         allow_reorder = False,
         widget = DataGridWidget(
             label='Form fields to Payables mapping',
             label_msgid = "label_getpaid_field_map",
             description="""Associate the paths of form fields with payables.""",
             description_msgid = 'help_getpaid_field_map',
             columns= {
                 "field_path" : FixedColumn("Form Fields (path)", visible=False),
                 "form_field" : FixedColumn("Form Fields"),
                 "payable_path" : SelectColumn("Payables", 
                                           vocabulary="buildPayablesList")
             },
             i18n_domain = "getpaid.formgen",
             ),
        ),
    StringField('GPMakePaymentButton',
                default='Make Payment',
                 mutator='setGPSubmit',
                 widget=StringWidget(
                        label=u'Get Paid Payment Button Label',
                        i18n_domain = "getpaidpfgadapter",
                        label_msgid = "label_getpaid_button_label",
                        ))
 
    
))

finalizeATCTSchema(schema, folderish=False, moveDiscussion=False)

class BillingInfo( getpaid.core.options.PropertyBag ):
    title = "Billing Information"

    def __init__(self, context):
        # need a context to be able to get the current available credit
        # cards. purge state afterwards
        self.context = context

    def __getstate__( self ):
        # don't store persistently
        raise RuntimeError("Storage Not Allowed")

BillingInfo = BillingInfo.makeclass( getpaid.core.interfaces.IUserPaymentInformation )

def order_fields(x,y):
    if x[1]>y[1]:
        return 1
    if x[1]<y[1]:
        return -1
    else:
        return 0


def getAvailableCreditCards(self):
    """
    We need the vocabulary for the credit cards in a form that can be understood by pfg
    """
    portal = zapi.getSiteManager()
    portal = getToolByName(portal,'portal_url').getPortalObject()
    manage_options = IGetPaidManagementOptions(portal)
    credit_cards = manage_options.accepted_credit_cards
    available_cards = []
    for card in credit_cards:
        available_cards.append("%s|%s" % (card,card) )
    return tuple(available_cards)
    

class GetpaidPFGAdapter(FormActionAdapter):
    """
    Do PloneGetPaid stuff upon PFG submit.
    """
    schema = schema
    security = ClassSecurityInfo()

    fieldsetType = atapi.ATFieldProperty('GPFieldsetType')    
    makePaymentButton = atapi.ATFieldProperty('GPMakePaymentButton')
    
    portal_type = 'GetpaidPFGAdapter'
    archetype_name = 'Getpaid Adapter'
    content_icon = 'getpaid.gif'

    available_templates = {'One Page Checkout': '_one_page_checkout_init',
                           'Multi item cart add': '_multi_item_cart_add_init' }
    success_callback = None
    
    checkout_fields = {'User Data':[{
                       
        'name':['FormStringField',{'title':u"Your Full Name",
                                   'required':True},10],
        'phone_number':['FormStringField',{'title':u"Phone Number",
                                         'description':u"Only digits allowed - e.g. 3334445555 and not 333-444-5555",
                                           'required':True},11],
        'email':['FormStringField',{'title':u"Email",
                                  'description':u"",
                                    'required':True},11]
        },10],
                       'Billing Address Info':[{
        'bill_first_line':['FormStringField',{'title':u"Address 1",
                                              'required':True},10],
        'bill_second_line':['FormStringField',{'title':u"Address 2"},11],
        'bill_city':['FormStringField',{'title':u"City",
                                        'required':True},12],
        'bill_postal_code':['FormStringField',{'title':u"Zip/Postal Code",
                                               'required':True},13],
        'bill_country':['FormStringField',{'title':u"Country Code",
                                           'description':'Your Country ISO code',
                                               'required':True},14],
        'bill_state':['FormStringField',{'title':u"State Code",
                                         'description':'Your State ISO code',
                                               'required':True},15]
        },20],
                        'Credit Info':[{
        'credit_card_type':['FormSelectionField',{'title':u"Credit Card Type",
                                                  'fgVocabulary':getAvailableCreditCards,
                                                  'required':True},10],
        'name_on_card':['FormStringField',{'title':u"Card Holder Name",
                                         'description':u"Enter the full name, as it appears on the card.",
                                           'required':True},11],
        'credit_card':['FormStringField',{'title':u"Credit Card Number",
                                          'description':u"Only digits allowed - e.g. 4444555566667777 and not 4444-5555-6666-7777 ",
                                          'required':True},12],
        'cc_expiration':['FormDateField',{'title':u"Credit Card Expiration Date",
                                          'required':True},13],
        'cc_cvc':['FormIntegerField',{'title':u"Credit Card Verfication Number",
                                      'description':u"For MC, Visa, and DC, this is a 3-digit number on back of the card.  For AmEx, this is a 4-digit code on front of card. ",
                                      'fgmaxlength':4,
                                      'required':True},14],
        },30]
        
                       }

    



    #--------------------------------------------------------------------------#
    #One page checkout methods
    #--------------------------------------------------------------------------#
    def _one_page_checkout_init(self):
        """
        We add all the required fields for getpaid checkout
        TODO: Find a way to get ordered fields
        """
        oids = self.keys()
        ordered_keys = [(afieldset,self.checkout_fields[afieldset][1]) for afieldset in self.checkout_fields.keys()]
        ordered_keys.sort(order_fields)
        ordered_keys = [ordered_key[0] for ordered_key in ordered_keys] # just lazyness to refactor the rest of the method
        
        for frameset in ordered_keys:
            if frameset in oids:
                return
            self.invokeFactory('FieldsetFolder',frameset)
            frame_folder = self[frameset]
            frame_folder.setTitle(frameset)
            ordered_fields = [(fld,self.checkout_fields[frameset][0][fld][2]) for fld in self.checkout_fields[frameset][0].keys()]
            ordered_fields.sort(order_fields)
            ordered_fields = [ordered_key[0] for ordered_key in ordered_fields]
            fs_oids = frame_folder.keys()
            
            for field in ordered_fields:
                if not field in fs_oids:
                    aField = self.checkout_fields[frameset][0][field]
                    frame_folder.invokeFactory(aField[0],field)
                    obj = frame_folder[field]
                    obj.fgField.__name__ = field
                    attribute_list = aField[1]
                    for attr in attribute_list.keys():
                        if hasattr(obj,"set%s" % attr.capitalize()):
                            #A little dirty but apparently calling the method does more
                            #than just setting the value (I will use my it's 2 AM card here)
                            if callable(attribute_list[attr]):
                                getattr(obj,"set%s" % attr.capitalize())(attribute_list[attr](self))
                            else:
                                getattr(obj,"set%s" % attr.capitalize())(attribute_list[attr])                                
                        else:
                            if callable(attribute_list[attr]):
                                obj.getField(attr).set(obj,attribute_list[attr](self))
                            else:
                                obj.getField(attr).set(obj,attribute_list[attr])
                    if 'required' in attribute_list:
                        obj.fgField.required = True

        self.success_callback = "_one_page_checkout_success"

    def _one_page_checkout_success(self, fields, REQUEST=None):
        """
        this process is quite like the regular one except it will use a disposable cart
        """
        portal = zapi.getSiteManager()
        portal = getToolByName(portal,'portal_url').getPortalObject()
        adapters = self.getSchemaAdapters()
        cartKey = "multishot:%s" % aq_parent(self).title
        shopping_cart = component.getUtility(GPInterfaces.IShoppingCartUtility).get(portal, key=cartKey)
        
        portal_catalog = getToolByName(self, 'portal_catalog')
        #The split is because of the fields inside fieldsets that are presented as a "fieldset,field" string
        form_payable = dict((p['field_path'].split(',')[-1], p['payable_path']) for p in self.payablesMap if p['payable_path'])
        parent_node = self.getParentNode()
        has_products = 0
        error_fields = {}
        for field in fields:
            if field.getId() in form_payable:
                try:

                    content = parent_node.unrestrictedTraverse(form_payable[field.getId()], None)                        

                    if content is not None:

                        if IVariableAmountDonatableMarker.providedBy(content):
                            arg = float(REQUEST.form.get(field.fgField.getName()))
                        else:
                            arg = int(REQUEST.form.get(field.fgField.getName()))

                        if arg > 0:
                            has_products += 1
                            try:
                                item_factory = zope.component.getMultiAdapter((shopping_cart, content),
                                    getpaid.core.interfaces.ILineItemFactory)
                                item_factory.create(arg)
                            except zope.component.ComponentLookupError, e:
                                pass
                        elif arg < 0 :
                            error_fields[field.getId()] = "The value for this field is not allowed"
                except KeyError, e:
                    pass
                except ValueError, e:
                    error_fields[field.getId()] = "The value for this field is not allowed"
            else:
                for adapter in adapters.values():
                    adapter_fields = zope.schema.getFields(adapter.schema)
                    if field.getId() in adapter_fields.keys():
                        setattr(adapter,field.getId(),REQUEST.form.get(field.fgField.getName()))
        
        if error_fields:
            error_fields.update({FORM_ERROR_MARKER:'Some of the values where incorrect'})
            return error_fields
        if has_products == 0:
            return {FORM_ERROR_MARKER:'There are no products in the order'}
        
        checkout_process = MakePaymentProcess(portal, adapters) 
        result = checkout_process(shopping_cart)

        # I only want this cart to stick around for this submission of the form
        # so destroy it to make sure it doesn't hang around
        scu = component.getUtility(GPInterfaces.IShoppingCartUtility)
        scu.destroy(portal, key=cartKey)

        if result:
            return {FORM_ERROR_MARKER:'%s' % result}
        REQUEST.response.redirect(self.getNextURL(checkout_process.order, portal))
            
        
    #--------------------------------------------------------------------------#
    #Multi item cart add methods
    #--------------------------------------------------------------------------#

    def _multi_item_cart_add_init(self):
        self.success_callback = "_multi_item_cart_add_success"        

                    
    def _multi_item_cart_add_success(self, fields, REQUEST=None):
        scu = zope.component.getUtility(getpaid.core.interfaces.IShoppingCartUtility)
        cart = scu.get(self, create=True)
        portal_catalog = getToolByName(self, 'portal_catalog')
        form_payable = dict((p['field_path'], p['payable_path']) for p in self.payablesMap if p['payable_path'])
        parent_node = self.getParentNode()

        formFolder = aq_parent(self)
        formFolderPath = formFolder.getPhysicalPath()
        for field in fields:
            fieldId = ",".join(field.getPhysicalPath()[len(formFolderPath):])

            if fieldId in form_payable:
                try:
                    content = parent_node.unrestrictedTraverse(form_payable[fieldId], None)

                    if content is not None:
                        if IVariableAmountDonatableMarker.providedBy(content):
                            arg = float(REQUEST.form.get(field.fgField.getName()))
                        else:
                            arg = int(REQUEST.form.get(field.fgField.getName()))

                        if arg > 0:
                            try:
                                item_factory = zope.component.getMultiAdapter((cart, content),
                                    getpaid.core.interfaces.ILineItemFactory)
                                item_factory.create(arg)
                            except zope.component.ComponentLookupError, e:
                                pass
                except KeyError, e:
                    pass
                except ValueError, e:
                    pass

    #--------------------------------------------------------------------------#
    #Helper methods
    #--------------------------------------------------------------------------#
    
    def getSchemaAdapters(self):
        adapters = {}
        portal = zapi.getSiteManager()
        portal = getToolByName(portal,'portal_url').getPortalObject()
        user = getSecurityManager().getUser()
        formSchemas = component.getUtility(GPInterfaces.IFormSchemas)
        #persistent sections
        for section in ('contact_information', 'billing_address', 'shipping_address'):
            interface = formSchemas.getInterface(section)
            adapter = component.queryAdapter(user,interface)
            if adapter is None:
                adapter = formSchemas.getBagClass(section)()
            adapters[interface]=adapter
        #non persistent sections
        adapters[formSchemas.getInterface('payment')] = formSchemas.getBagClass('payment')(portal)
        return adapters

    def onSuccess(self, fields, REQUEST=None):
        """
        Will call the on success method according to the chosen template
        """

        result = getattr(self,self.success_callback)(fields, REQUEST)

        # This needs to occur after we add the item to our cart since
        # doing that sets came_from_url to the item
        if getattr(self, 'useFormAsContinueDestination', False):
            sessions.set_came_from_url(aq_parent(self))

        return result
    
    def setGPTemplate(self, template):
        """
        This sets the template for each kind of form
        """
        if template:
            self.getField('GPFieldsetType').set(self,template)
            getattr(self,self.available_templates[template])()

    def setGPSubmit(self, submit_legend):
        """
        This sets the chosen submit button legend
        """
        if submit_legend:
            self.makePaymentButton = submit_legend
            self.setSubmitLabel(submit_legend)

    def updateFieldsData(self, fieldData):
        for fields in fieldData:
            if not 'payable_path' in fields.keys():
                return

            if fields['payable_path']:
                #There should not be more than one level of recursion
                local_widget_name = fields['field_path'].split("-->")
                if len(local_widget_name) > 1:
                    widget = self[local_widget_name[0]][local_widget_name[1]]
                else:
                    widget = self[local_widget_name[0]]
                    
                parent_node = self.getParentNode()
                content = parent_node.restrictedTraverse(fields['payable_path'], None)
                
                payable = GPInterfaces.IPayable(content)    
                if payable.price:
                    title = content.title + " : $%0.2f" % (payable.price)
                else:
                    title = content.title
                description = content.description

                widget.setTitle(title)
                widget.setDescription(description)
                
        
    security.declareProtected(ModifyPortalContent, 'generateFormFieldRows')
    def generateFormFieldRows(self):
        """This method returns a list of rows for the field mapping
           ui. One row is returned for each field in the form folder.
        """
        fixedRows = []
        
        for formFieldTitle, formFieldPath in self._getIPloneFormGenFieldsPathTitlePair():
            logger.debug("creating mapper row for %s" % formFieldTitle)
            fixedRows.append(FixedRow(keyColumn="form_field",
                                      initialData={"form_field" : formFieldTitle, 
                                                   "field_path" : formFieldPath,
                                                   "sf_field" : ""}))
        return fixedRows

    def _getIPloneFormGenFieldsPathTitlePair(self):
        formFolder = aq_parent(self)
        formFolderPath = formFolder.getPhysicalPath()
        formFieldTitles = []
        
        for formField in formFolder.objectIds():
            fieldObj = getattr(formFolder, formField)
            if IPloneFormGenField.providedBy(fieldObj):
                formFieldTitles.append((fieldObj.Title().strip(),
                                        ",".join(fieldObj.getPhysicalPath()[len(formFolderPath):])))
        
            # can we also inspect further down the chain
            if fieldObj.isPrincipiaFolderish:
                # since nested folders only go 1 level deep
                # a non-recursive approach approach will work here
                for subFormField in fieldObj.objectIds():
                    subFieldObj = getattr(fieldObj, subFormField)
                    if IPloneFormGenField.providedBy(subFieldObj):
                        # we append a list in this case
                        formFieldTitles.append(("%s --> %s" % (fieldObj.Title().strip(),
                                                               subFieldObj.Title().strip()),
                                                ",".join(subFieldObj.getPhysicalPath()[len(formFolderPath):])))
        
        return formFieldTitles

    def getNextURL(self, order, context):
        """
        Borrowed from GetPaid this will redirect on submit according to the
        result of the operation
        """
        state = order.finance_state
        f_states = GPInterfaces.workflow_states.order.finance
        base_url = context.absolute_url()
        if not 'http://' in base_url:
            base_url = base_url.replace("https://", "http://")

        if state in (f_states.CANCELLED,
                     f_states.CANCELLED_BY_PROCESSOR,
                     f_states.PAYMENT_DECLINED):
            return base_url + '/@@getpaid-cancelled-declined'
        
        if state in (f_states.CHARGEABLE,
                     f_states.CHARGING,
                     f_states.REVIEWING,
                     f_states.CHARGED):
            return base_url + '/@@getpaid-thank-you?order_id=%s&finance_state=%s' %(order.order_id, state)

    def getAvailableGetPaidForms(self):
        """
        We will provide a 'vocabulary' with the predefined form templates available
        It is not possible for the moment to do any kind of form without restriction
        """
        
        available_template_list = DisplayList()
        for field in self.available_templates.keys():
            available_template_list.add( field, field )
        return available_template_list


    def buildPayablesList(self):
        """
        Creates a list with the payable marked items on our store
        """
        portal_catalog = getToolByName(self, 'portal_catalog')
        portal_url = getToolByName(self, 'portal_url')
        portal_path = '/'.join(portal_url.getPortalObject().getPhysicalPath())
        buyables = portal_catalog.searchResults(
            dict(object_provides='Products.PloneGetPaid.interfaces.IPayableMarker',
                 path=portal_path))
        stuff = [('', '')]
        for b in buyables:
            o = b.getObject()
            payable = GPInterfaces.IPayable(o)
            if payable.price:
                stuff.append((b.getPath(), o.title + " : %0.2f" % (payable.price)))
            else:
                stuff.append((b.getPath(), o.title))
        display = DisplayList(stuff)
        return display

    
atapi.registerType(GetpaidPFGAdapter, PROJECTNAME)




"""Common configuration constants
"""
GLOBALS = globals()
product_globals = GLOBALS

from Products.CMFCore.permissions import setDefaultRoles

PROJECTNAME = 'getpaid.formgen'

DEPENDENCIES = []
PRODUCT_DEPENDENCIES = []
ADD_PERMISSIONS = {
    'GetpaidPFGAdapter' : 'PloneFormGen: Add GetPaid adapter',
    'GetPaidFormMailerAdapter' : 'PloneFormGen: Add GetPaid Mailer adapter',
}
setDefaultRoles(ADD_PERMISSIONS['GetpaidPFGAdapter'], ['Manager',])
setDefaultRoles(ADD_PERMISSIONS['GetPaidFormMailerAdapter'], ['Manager',])



"""
 A form action adapter that e-mails input.
"""

__author__  = 'Rob LaRubbio <rob@onenw.org>'
__docformat__ = 'plaintext'


########################
# The formMailerAdapter code and schema borrow heavily
# from PloneFormGen <http://plone.org/products/ploneformgen>
# by Steve McMahon <steve@dcn.org> which borrows heavily 
# from PloneFormMailer <http://plone.org/products/ploneformmailer>
# by Jens Klein and Reinout van Rees.
#
# Author:       Jens Klein <jens.klein@jensquadrat.com>
#
# Copyright:    (c) 2004 by jens quadrat, Klein & Partner KEG, Austria
# Licence:      GNU General Public Licence (GPL) Version 2 or later
#######################

import logging

from cPickle import loads, dumps

from AccessControl import ClassSecurityInfo

from Acquisition import aq_parent

from Products.Archetypes.public import *
from Products.Archetypes.utils import OrderedDict
from Products.Archetypes.utils import shasattr

from Products.ATContentTypes.content.schemata import finalizeATCTSchema
from Products.ATContentTypes.content.base import registerATCT

from Products.CMFCore.permissions import View, ModifyPortalContent

from Products.Archetypes.utils import DisplayList

from Products.PloneFormGen import dollarReplace

from Products.PloneFormGen.content.fields import *
from Products.PloneFormGen.content.ya_gpg import gpg

from Products.PloneFormGen.content.formMailerAdapter import FormMailerAdapter, formMailerAdapterSchema

from getpaid.formgen.config import PROJECTNAME

from email.Header import Header

# Get Paid events
import zope
from getpaid.core.interfaces import workflow_states, IShoppingCartUtility, IShippableOrder, IShippingRateService, IShippableLineItem
from zope.app.component.hooks import getSite
from zope.app.annotation.interfaces import IAnnotations

import copy
from Acquisition import aq_base

logger = logging.getLogger("GetPaidFormMailer")

getPaidFormMailerAdapterSchema = formMailerAdapterSchema.copy()


class GetPaidFormMailerAdapter(FormMailerAdapter):
    """ A form action adapter that will e-mail form input. """

    schema = getPaidFormMailerAdapterSchema
    portal_type = meta_type = 'GetPaidFormMailerAdapter'
    archetype_name = 'GetPaid Mailer Adapter'
    content_icon = 'mailaction.gif'

    security       = ClassSecurityInfo()

    security.declarePrivate('onSuccess')
    def onSuccess(self, fields, REQUEST=None):
        """
        e-mails data.
        """
        all_fields = [f for f in fields
            if not (f.isLabel() or f.isFileField()) and not (getattr(self, 'showAll', True) and f.getServerSide())]

        # which form fields should we show?
        if getattr(self, 'showAll', True):
            live_fields = all_fields 
        else:
            live_fields = \
                [f for f in all_fields
                   if f.fgField.getName() in getattr(self, 'showFields', ())]

        if not getattr(self, 'includeEmpties', True):
            all_fields = live_fields
            live_fields = []
            for f in all_fields:
                value = f.htmlValue(request)
                if value and value != 'No Input':
                    live_fields.append(f)
                
        formFields = []
        for field in live_fields:
            formFields.append( (field.title, field.htmlValue(REQUEST)) )

        scu = zope.component.getUtility(IShoppingCartUtility)

        # I need to figure out which cart to get.  The default, or multishot
        # to do that I find the first getpaid adapter (if there are multiple 
        # then this will fail).  I then look for the attr success_callback
        formFolder = aq_parent(self)
        adapter = formFolder.objectValues('GetpaidPFGAdapter')[0]
        cart = None
        if (adapter.success_callback == '_one_page_checkout_success'):
            cartKey = "multishot:%s" % formFolder.title
            cart = scu.get(self, key=cartKey)
        else:
            cart = scu.get(self, create=True)

        if (cart == None):
            logger.info("Unable to get cart")
        else:
            annotation = IAnnotations(cart)

            if "getpaid.formgen.mailer.adapters" in annotation:
                adapters = annotation["getpaid.formgen.mailer.adapters"]

                if not self.title in adapters:
                    adapters.append(self.title)
            else:
                adapters = [self.title]

            annotation["getpaid.formgen.mailer.adapters"] = adapters

            annotationKey = "getpaid.formgen.mailer.%s" % self.title

            data = {}
            data['formFields'] = formFields

            attachments = self.get_form_attachments(fields, REQUEST)
            data['attachments'] = attachments

            # This is a complete hack and I can't believe it isn't
            # going to come back and break in the future
            # I just need some of the request for my supers
            # implementation.  I can't pickle an aq wrapped object, so
            # I resort to this.  Perhaps it's better to reimplement
            # the code from my super I'm reusing? 
            # (note I did end up pulling more code into this class)
            req = {}
            req['form'] = {}
            # ignore any form element that is a file upload field
            for key in REQUEST.form.keys():
                value = REQUEST.form[key]
                if value and not isinstance(value, FileUpload):
                    req['form'][key] = value
                
            for key in getattr(self, 'xinfo_headers', []):
                if REQUEST.has_key(key):
                    req[key] = REQUEST[key]

            data['request'] = req

            # adapter is attached to the main ZODB
            # If using mount points then storage of adapter fails
            # due to cross db commit. We pickle to get a 
            # clean copy to store.
            data['adapter'] = loads( dumps( aq_base(self) ) )

            annotation[annotationKey] = data

    security.declarePrivate('_dreplace')
    def _dreplace(self, s):
        _form = getattr(self.REQUEST, 'form', None)
        if _form is None:
            _form = self.REQUEST['form']

        return dollarReplace.DollarVarReplacer(_form).sub(s)

    security.declarePrivate('getMailBodyDefault')
    def getMailBodyDefault(self):
        """ Get default mail body from our tool """
        
        return DEFAULT_MAILTEMPLATE_BODY

    # Todo implement this to pull attachments out of the annotation
    def get_attachments(self, fields, request):
        """Return all attachments that were uploaded in form
           and stored in an annotation
        """

        scu = zope.component.getUtility(IShoppingCartUtility)
        cart = scu.get(self, create=True)

        attachments = []
        if (cart == None):
            logger.info("Unable to get cart")
        else:
            annotation = IAnnotations(cart)

            annotationKey = "getpaid.formgen.mailer.%s" % self.title
            if annotationKey in annotation:
                data = annotation[annotationKey]
                attachments = data["attachments"]

        return attachments

    def get_form_attachments(self, fields, request):
        """Return all attachments uploaded in form.
        """

        from ZPublisher.HTTPRequest import FileUpload

        attachments = []

        for field in fields:
            if field.isFileField():
                file = request.form.get('%s_file' % field.__name__, None)
                if file and isinstance(file, FileUpload) and file.filename != '':
                    file.seek(0) # rewind
                    data = file.read()
                    filename = file.filename
                    mimetype, enc = guess_content_type(filename, data, None)
                    attachments.append((filename, mimetype, enc, data))
        return attachments

    security.declarePrivate('get_mail_body')
    def get_mail_body(self, fields, **kwargs):
        """Returns the mail-body with footer.
        """

        bodyfield = self.getField('body_pt')
        
        # pass both the bare_fields (fgFields only) and full fields.
        # bare_fields for compatability with older templates,
        # full fields to enable access to htmlValue
        body = bodyfield.get(self, formFields=fields, **kwargs)

        if isinstance(body, unicode):
            body = body.encode(self.getCharset())

        keyid = getattr(self, 'gpg_keyid', None)
        encryption = gpg and keyid

        if encryption:
            bodygpg = gpg.encrypt(body, keyid)
            if bodygpg.strip():
                body = bodygpg

        return body

    # I override this method since I don't have a real request object.
    # I have a dict instead.
    security.declarePrivate('get_header_body_tuple')
    def get_header_body_tuple(self, fields, request,
                              from_addr=None, to_addr=None,
                              subject=None, **kwargs):
        """Return header and body of e-mail as an 3-tuple:
        (header, additional_header, body)

        header is a dictionary, additional header is a list, body is a StringIO

        Keyword arguments:
        request -- (optional) alternate request object to use
        """

        body = self.get_mail_body(fields, **kwargs)

        # fields = self.fgFields()

        # get Reply-To
        reply_addr = None
        if shasattr(self, 'replyto_field'):
            reply_addr = request['form'].get(self.replyto_field, None)

        # get subject header
        nosubject = '(no subject)'
        if shasattr(self, 'subjectOverride') and self.getRawSubjectOverride():
            # subject has a TALES override
            subject = self.getSubjectOverride().strip()
        else:
            subject = getattr(self, 'msg_subject', nosubject)
            subjectField = request['form'].get(self.subject_field, None)
            if subjectField is not None:
                subject = subjectField
            else:
                # we only do subject expansion if there's no field chosen
                subject = dollarReplace.DollarVarReplacer(request['form']).sub(subject)

        # Get From address
        if shasattr(self, 'senderOverride') and self.getRawSenderOverride():
            from_addr = self.getSenderOverride().strip()
        else:
            pprops = getToolByName(self, 'portal_properties')
            site_props = getToolByName(pprops, 'site_properties')
            portal = getToolByName(self, 'portal_url').getPortalObject()
            from_addr = from_addr or site_props.getProperty('email_from_address') or \
                        portal.getProperty('email_from_address')

        # Get To address and full name
        if shasattr(self, 'recipientOverride') and self.getRawRecipientOverride():
            recip_email = self.getRecipientOverride()
        else:
            recip_email = None
            if shasattr(self, 'to_field'):
                recip_email = request['form'].get(self.to_field, None)
            if not recip_email:
                recip_email = self.recipient_email
        recip_email = self._destFormat( recip_email )

        recip_name = self.recipient_name.encode('utf-8')

        # if no to_addr and no recip_email specified, use owner adress if possible.
        # if not, fall back to portal email_from_address.
        # if still no destination, raise an assertion exception.
        if not recip_email and not to_addr:
            ownerinfo = self.getOwner()
            ownerid=ownerinfo.getId()
            pms = getToolByName(self, 'portal_membership')
            userdest = pms.getMemberById(ownerid)
            toemail = userdest.getProperty('email', '')
            if not toemail:
                portal = getToolByName(self, 'portal_url').getPortalObject()
                toemail = portal.getProperty('email_from_address')                
            assert toemail, """
                    Unable to mail form input because no recipient address has been specified.
                    Please check the recipient settings of the PloneFormGen "Mailer" within the
                    current form folder.
                """
            to = '%s <%s>' %(ownerid,toemail)
        else:
            to = to_addr or '%s %s' % (recip_name, recip_email)

        headerinfo = OrderedDict()

        headerinfo['To'] = self.secure_header_line(to)
        headerinfo['From'] = self.secure_header_line(from_addr)
        if reply_addr:
            headerinfo['Reply-To'] = self.secure_header_line(reply_addr)

        # transform subject into mail header encoded string
        portal = getToolByName(self, 'portal_url').getPortalObject()
        email_charset = portal.getProperty('email_charset', 'utf-8')
        msgSubject = self.secure_header_line(subject).encode(email_charset, 'replace')
        msgSubject = str(Header(msgSubject, email_charset))
        headerinfo['Subject'] = msgSubject

        headerinfo['MIME-Version'] = '1.0'

        # CC
        cc_recips = filter(None, self.cc_recipients)
        if cc_recips:
            headerinfo['Cc'] = self._destFormat( cc_recips )

        # BCC
        bcc_recips = filter(None, self.bcc_recipients)
        if shasattr(self, 'bccOverride') and self.getRawBccOverride():
            bcc_recips = self.getBccOverride()
        if bcc_recips:
            headerinfo['Bcc'] = self._destFormat( bcc_recips )

        for key in getattr(self, 'xinfo_headers', []):
            headerinfo['X-%s' % key] = self.secure_header_line(request.get(key, 'MISSING'))

        # return 3-Tuple
        return (headerinfo, self.additional_headers, body)

    security.declareProtected(View, 'allFieldDisplayList')
    def allFieldDisplayList(self):
        """ returns a DisplayList of all fields """

        ret = []
        for field in self.fgFieldsDisplayList():
            ret.append(field)

        return ret


    def fieldsDisplayList(self):
        """ returns display list of fields with simple values """

        ret = []

        foo = self.fgFieldsDisplayList(
            withNone=True,
            noneValue='#NONE#',
            objTypes=(
                'FormSelectionField',
                'FormStringField',
                )
            )
        for field in foo:
            ret.append(field)

        ret.append(EMAIL)

        return ret

    security.declareProtected(ModifyPortalContent, 'setShowFields')
    def setShowFields(self, value, **kw):
        """ Reorder form input to match field order """
        # This wouldn't be desirable if the PickWidget
        # retained order.

        self.showFields = []
        for field in self.fgFields(excludeServerSide=False):
            id = field.getName()
            if id in value:
                self.showFields.append(id)


registerATCT(GetPaidFormMailerAdapter, PROJECTNAME)

def handleOrderWorkflowTransition( order, event ):

    try:
        if order.finance_state == event.destination and event.destination == workflow_states.order.finance.CHARGED:
            annotation = IAnnotations(order.shopping_cart)
            
            if "getpaid.formgen.mailer.adapters" in annotation:
                adapters = annotation["getpaid.formgen.mailer.adapters"]
                
                getPaidFields = _getValuesFromOrder(order)
                site = zope.app.component.hooks.getSite()
                for a in adapters:
                    annotationKey = "getpaid.formgen.mailer.%s" % a
                    data = annotation[annotationKey]
                    
                    formFields = data['formFields']
                    attachments = data['attachments']
                    request = data['request']
                    adapter = data['adapter']
                
                    adapter.__of__(site).send_form(formFields, request, getPaidFields=getPaidFields)
    except:
        # I catch evrything since an uncaught exception here will prevent
        # the order from moving to charged even though the charge 
        # has gone through
        logger.error("Exception sending email for order %s" % order.order_id)
        

def _getValuesFromOrder(order):
    ret = {}
    
    ret[NAME] = order.contact_information.name
    ret[PHONE_NUMBER] = order.contact_information.phone_number
    ret[EMAIL] = order.contact_information.email
    ret[CONTACT_ALLOWED] = order.contact_information.marketing_preference
    ret[EMAIL_PREFERENCE] = order.contact_information.email_html_format
    ret[BILLING_NAME] = order.billing_address.bill_name
    ret[BILLING_ORGANIZATION] = order.billing_address.bill_organization
    ret[BILLING_STREET_1] = order.billing_address.bill_first_line
    ret[BILLING_STREET_2] = order.billing_address.bill_second_line
    ret[BILLING_CITY] = order.billing_address.bill_city
    ret[BILLING_COUNTRY] = order.billing_address.bill_country
    ret[BILLING_STATE] = order.billing_address.bill_state    
    ret[BILLING_ZIP] = order.billing_address.bill_postal_code      
    ret[BILLING_PHONE] = order.bill_phone_number      
    
    if order.shipping_address.ship_same_billing:
        ret[SHIPPING_NAME] = order.billing_address.bill_name
        ret[SHIPPING_ORGANIZATION] = order.billing_address.bill_organization
        ret[SHIPPING_STREET_1] = order.billing_address.bill_first_line
        ret[SHIPPING_STREET_2] = order.billing_address.bill_second_line
        ret[SHIPPING_CITY] = order.billing_address.bill_city    
        ret[SHIPPING_COUNTRY] = order.billing_address.bill_country
        ret[SHIPPING_STATE] = order.billing_address.bill_state      
        ret[SHIPPING_ZIP] = order.billing_address.bill_postal_code         
    else:
        ret[SHIPPING_NAME] = order.shipping_address.ship_name
        ret[SHIPPING_ORGANIZATION] = order.shipping_address.ship_organization
        ret[SHIPPING_STREET_1] = order.shipping_address.ship_first_line
        ret[SHIPPING_STREET_2] = order.shipping_address.ship_second_line
        ret[SHIPPING_CITY] = order.shipping_address.ship_city    
        ret[SHIPPING_COUNTRY] = order.shipping_address.ship_country
        ret[SHIPPING_STATE] = order.shipping_address.ship_state      
        ret[SHIPPING_ZIP] = order.shipping_address.ship_postal_code         

    ret[SHIPPING_SERVICE] = getShippingService(order)
    ret[SHIPPING_METHOD] = getShippingMethod(order)
    ret[SHIPPING_WEIGHT] = getShipmentWeight(order)

    ret[ORDER_ID] = order.order_id
    ret[ORDER_DATE] = order.creation_date.ctime()
#    ret[ORDER_TAX_TOTAL] = order.getTaxCost()
    ret[ORDER_SHIPPING_TOTAL] = order.getShippingCost()
    ret[ORDER_SUB_TOTAL] = order.getSubTotalPrice()
    ret[ORDER_TOTAL] = order.getTotalPrice()
    ret[ORDER_TRANSACTION_ID] = order.user_payment_info_trans_id
    ret[CC_NAME] = order.name_on_card
    ret[CC_LAST_4] = order.user_payment_info_last4
    ret[ORDER_ITEMS_ARRAY] = []

    for item in order.shopping_cart.items():
        itemDict = {}
        itemDict[ITEM_QTY] = item[1].quantity
        itemDict[ITEM_ID] = item[1].item_id
        itemDict[ITEM_NAME] = item[1].name           
        itemDict[ITEM_PRODUCT_CODE] = item[1].product_code   
        itemDict[ITEM_COST] = item[1].cost           
        itemDict[ITEM_TOTAL_COST] = item[1].cost * item[1].quantity
        itemDict[ITEM_DESC] = item[1].description    

        # Check to see if a discount code was applied to this item
        itemDict[DISCOUNT_CODE] = ''
        itemDict[DISCOUNT_TITLE] = ''
        itemDict[DISCOUNT_TOTAL] = ''
        annotation = IAnnotations(item[1])
        if "getpaid.discount.code" in annotation:
            itemDict[DISCOUNT_CODE] = annotation["getpaid.discount.code"]
            itemDict[DISCOUNT_TITLE] = annotation["getpaid.discount.code.title"]
            itemDict[DISCOUNT_TOTAL] = annotation["getpaid.discount.code.discount"]

        ret[ORDER_ITEMS_ARRAY].append(itemDict)

    return ret
    
def getShippingService(order):
    if not hasattr(order,"shipping_service"):
        return None
    infos = order.shipping_service
    if infos:
        return infos

def getShippingMethod(order):
    # check the traversable wrrapper
    if not IShippableOrder.providedBy( order ):
        return None
    
    service = zope.component.queryUtility( IShippingRateService,
                                           order.shipping_service )
    
    # play nice if the a shipping method is removed from the store
    if not service: 
        return None
        
    return service.getMethodName( order.shipping_method )
    
def getShipmentWeight(order):
    """
    Lets return the weight in lbs for the moment
    """
    # check the traversable wrrapper
    if not IShippableOrder.providedBy( order ):
        return None

    totalShipmentWeight = 0
    for eachProduct in order.shopping_cart.values():
        if IShippableLineItem.providedBy( eachProduct ):
            weightValue = eachProduct.weight * eachProduct.quantity
            totalShipmentWeight += weightValue
    return totalShipmentWeight
       

NAME                  = u'Name'
PHONE_NUMBER          = u'Phone Number' 
EMAIL                 = u'Email'
CONTACT_ALLOWED       = u'Contact Allowed' 
EMAIL_PREFERENCE      = u'Email Format Preference'
BILLING_NAME          = u'Billing Address Name'
BILLING_ORGANIZATION  = u'Billing Organization'
BILLING_STREET_1      = u'Billing Address Street 1'
BILLING_STREET_2      = u'Billing Address Street 2'
BILLING_CITY          = u'Billing Address City'
BILLING_COUNTRY       = u'Billing Address Country'
BILLING_STATE         = u'Billing Address State'
BILLING_ZIP           = u'Billing Address Zip'
BILLING_PHONE         = u'Billing Phone Number'
SHIPPING_NAME         = u'Shipping Address Name'
SHIPPING_ORGANIZATION = u'Shipping Organization'
SHIPPING_STREET_1     = u'Shipping Address Street 1'
SHIPPING_STREET_2     = u'Shipping Address Street 2'
SHIPPING_CITY         = u'Shipping Address City'
SHIPPING_COUNTRY      = u'Shipping Address Country'
SHIPPING_STATE        = u'Shipping Address State'
SHIPPING_ZIP          = u'Shipping Address Zip'
SHIPPING_SERVICE      = u'Shipping Service'
SHIPPING_METHOD       = u'Shipping Method'
SHIPPING_WEIGHT       = u'Shipping Weight'
ORDER_ID              = u'Order Id'
ORDER_DATE            = u'Order Creation Date'
ORDER_TAX_TOTAL       = u'Tax Total'
ORDER_SHIPPING_TOTAL  = u'Shipping Total'
ORDER_SUB_TOTAL       = u'Order Subtotal'
ORDER_TOTAL           = u'Order Total'
ORDER_TRANSACTION_ID  = u'Order Transaction Id'
CC_NAME               = u'Cardholder Name'
CC_LAST_4             = u'CC Last 4'
ORDER_ITEMS_ARRAY     = u'Items'
DISCOUNT_CODE         = u'Discount Code'
DISCOUNT_TITLE        = u'Discount Title'
DISCOUNT_TOTAL        = u'Discount Total'
ITEM_QTY              = u'Line Item Quantity'
ITEM_ID               = u'Line Item Id'
ITEM_NAME             = u'Line Item Name'
ITEM_PRODUCT_CODE     = u'Line Item Product Code'
ITEM_COST             = u'Line Item Item Cost'
ITEM_TOTAL_COST       = u'Total Line Item Cost'
ITEM_DESC             = u'Line Item Item Description'

GetPaidFields = (
    NAME,
    PHONE_NUMBER,
    EMAIL,
    CONTACT_ALLOWED,
    EMAIL_PREFERENCE,
    BILLING_NAME,
    BILLING_ORGANIZATION,
    BILLING_STREET_1,
    BILLING_STREET_2,
    BILLING_CITY,
    BILLING_COUNTRY,
    BILLING_STATE,
    BILLING_ZIP,
    SHIPPING_NAME,
    SHIPPING_ORGANIZATION,
    SHIPPING_STREET_1,
    SHIPPING_STREET_2,
    SHIPPING_CITY,
    SHIPPING_COUNTRY,
    SHIPPING_STATE,
    SHIPPING_ZIP,
    SHIPPING_SERVICE,
    SHIPPING_METHOD,
    SHIPPING_WEIGHT,
    ORDER_ID,
    ORDER_DATE,
    ORDER_TAX_TOTAL,
    ORDER_SHIPPING_TOTAL,
    ORDER_SUB_TOTAL,
    ORDER_TOTAL,
    ORDER_TRANSACTION_ID,
    DISCOUNT_CODE,
    DISCOUNT_TITLE,
    DISCOUNT_TOTAL,
    CC_LAST_4,
    ORDER_ITEMS_ARRAY,
    ITEM_QTY,
    ITEM_ID,
    ITEM_NAME,
    ITEM_PRODUCT_CODE,
    ITEM_COST,
    ITEM_TOTAL_COST,
    ITEM_DESC,
    )
    
DEFAULT_MAILTEMPLATE_BODY = \
"""<html xmlns="http://www.w3.org/1999/xhtml">

  <head><title></title></head>

  <body>
    <p tal:content="here/getBody_pre | nothing" />
    <dl>
      <tal:block repeat="field options/formFields">
        <dt tal:content="python:field[0]"/>
        <dt tal:content="python:field[1]"/>
      </tal:block>
    </dl>
    <tal:block define="field options/getPaidFields">
      <div>
        <div style="float:left; width:30%">
          <fieldset>
	    <legend> Billing Address </legend>
            <span tal:content="python:field['Billing Address Name']">Name</span><br />
            <span tal:content="python:field['Billing Organization']">Org</span><br />
            <span tal:content="python:field['Billing Address Street 1']">Street 1</span><br />
            <span tal:content="python:field['Billing Address Street 2']">Street 2</span><br />
            <span tal:content="python:field['Billing Address City']">City</span><br />
            <span tal:content="python:field['Billing Address Country']">Country</span><br />
            <span tal:content="python:field['Billing Address State']">State</span><br />
            <span tal:content="python:field['Billing Address Zip']">Zip</span><br />
          </fieldset>
        </div>
        <div style="float: left; padding-left: 3em; width: 30%;">
          <fieldset>
            <legend> Mailing Address </legend>
            <span tal:content="python:field['Shipping Address Name']">Name</span><br />
            <span tal:content="python:field['Shipping Organization']">Org</span><br />
            <span tal:content="python:field['Shipping Address Street 1']">Street 1</span><br />
            <span tal:content="python:field['Shipping Address Street 2']">Street 2</span><br />
            <span tal:content="python:field['Shipping Address City']">City</span><br />
            <span tal:content="python:field['Shipping Address Country']">Country</span><br />
            <span tal:content="python:field['Shipping Address State']">State</span><br />
            <span tal:content="python:field['Shipping Address Zip']">Zip</span><br />
          </fieldset>
        </div>
      </div>

      <div style="clear:both;"><!-- --></div>
      
      <div>
	<fieldset>
 	  <legend> Contact Information </legend>
          Email: <span tal:content="python:field['Email']">Email</span><br />
          Billing Phone: <span tal:content="python:field['Billing Phone Number']">Phone</span><br />
          Phone: <span tal:content="python:field['Phone Number']">Phone</span><br />
        </fieldset>
      </div>

      <div style="clear:both;"><!-- --></div>

      <div>
	<fieldset>
 	  <legend> Shipping Information </legend>
	  
          <table class="listing">
            <tbody>           
              <tr>
                <td>
                  Shipping Service
                </td>
                <td>
                  <span tal:content="python:field['Shipping Service']">Shipping Service</span>
                </td>
              </tr>
              <tr>
                <td>
                  Shipping Method
                </td>
                <td>
                  <span tal:content="python:field['Shipping Method']">Shipping Method</span>
                </td>
              </tr>
              <tr>
                <td>
                  Shipping Weight
                </td>
                <td>
                  <span tal:content="python:field['Shipping Weight']">Shipping Weight</span>
                </td>
              </tr>
            </tbody>
          </table>
	</fieldset>
      </div>

      <div style="clear:both;"><!-- --></div>

      <div class="cart-listing">
	<fieldset>
	  <legend> Shopping Cart </legend>
	  
	  <table class="listing">
	    <thead>
              <tr>             
		<th>
		  Quantity
		</th>
		<th>
		  Name
		</th>
		<th>
		  Price
		</th>
		<th>
		  Total
		</th>
              </tr>
	    </thead>

	    <tbody>           
              <tal:block tal:repeat="item python:field['Items']">
		<tr>
		  <td>
		    <span tal:content="python:item['Line Item Quantity']">1</span>
		  </td>
		  <td>
		    <span tal:content="python:item['Line Item Name']">Name</span>
		  </td>
		  <td>
		    <span tal:content="python:item['Line Item Item Cost']">Price</span>
		  </td>
		  <td>
		    <span tal:content="python:item['Total Line Item Cost']">Total Cost</span>
		  </td>
		</tr>
              </tal:block>
	    </tbody>
	  </table>

	  <div class="getpaid-totals">
	    <table class="listing">
              <tr>
		<th>SubTotal</th>
		<td><span tal:content="python:field['Order Subtotal']">Subtotal</span></td>
              </tr>
              <tr>
		<th>Shipping</th>
		<td><span tal:content="python:field['Shipping Total']">Shipping</span></td>
              </tr>
	      <!--
		  <tr>
		    <th>Tax</th>
		    <td><span tal:content="python:field['Tax']">Tax</span></td>
		  </tr>
		  -->
              <tr>
		<th>Total</th>
		<td><span tal:content="python:field['Order Total']">Order Total</span></td>
              </tr>
	    </table>
	  </div>
	</fieldset>
      </div>
    </tal:block>
    <p tal:content="here/getBody_post | nothing" />
    <pre tal:content="here/getBody_footer | nothing" />
  </body>
</html>
"""
